/* _NVRM_COPYRIGHT_BEGIN_
 *
 * Copyright 2001 by NVIDIA Corporation.  All rights reserved.  All
 * information contained herein is proprietary and confidential to NVIDIA
 * Corporation.  Any use, reproduction, or disclosure without the written
 * permission of NVIDIA Corporation is prohibited.
 *
 * _NVRM_COPYRIGHT_END_
 */


#ifndef _NV_LINUX_H_
#define _NV_LINUX_H_

#include "nv.h"

#include <linux/config.h>

#if defined (CONFIG_MODVERSIONS) && !defined (MODVERSIONS)
#  define MODVERSIONS
#endif

#if defined (MODVERSIONS)
#include <linux/modversions.h>
#endif

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/version.h>

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 2, 12)
#  error This driver does not support 2.2.11 or earlier kernels!
#elif LINUX_VERSION_CODE < KERNEL_VERSION(2, 3, 0)
#  define KERNEL_2_2
#  warning NVIDIA is considering dropping support for linux-2.2 
#  warning kernels. While end users are free to maintain their 
#  warning own patches, or stick with current drivers, our new 
#  warning drivers will not work "out of the box." If you are 
#  warning concerned about lack of support for 2.2 kernels, 
#  warning please let us know at linux-bugs@nvidia.com; we would
#  warning like to know how many users would be seriously 
#  warning impacted by this decision.
#elif LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 0)
#  error This driver does not support 2.3.x development kernels!
#elif LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
#  define KERNEL_2_4
#elif LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
#  error This driver does not support 2.5.x development kernels!
#  define KERNEL_2_5
#else
#  error This driver does not support 2.6.x or newer kernels!
#endif

#if defined (__ia64)
#  if LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 13)
#    error This driver does not support 2.4.12 or earlier kernels!
#  endif
#endif

#if defined (CONFIG_SMP) && !defined (__SMP__)
#define __SMP__
#endif

#include <linux/types.h>            /* pic_t, size_t, __u32, etc        */
#include <linux/errno.h>            /* error codes                      */
#include <linux/list.h>             /* circular linked list             */
#include <linux/stddef.h>           /* NULL, offsetof                   */
#include <linux/wait.h>             /* wait queues                      */
#include <linux/tqueue.h>           /* struct tq_struct                 */

#include <linux/slab.h>             /* kmalloc, kfree, etc              */
#include <linux/vmalloc.h>          /* vmalloc, vfree, etc              */

#include <linux/poll.h>             /* poll_wait                        */
#include <linux/delay.h>            /* mdelay, udelay                   */

#include <linux/pci.h>              /* pci_find_class, etc              */
#include <linux/wrapper.h>          /* mem_map_reserve                  */
#include <linux/interrupt.h>        /* mark_bh, init_bh, remove_bh      */
#include <linux/timer.h>

#include <asm/system.h>             /* cli, sli, save_flags             */
#include <asm/io.h>                 /* ioremap, virt_to_phys            */
#include <asm/uaccess.h>            /* access_ok                        */
#include <asm/page.h>               /* PAGE_OFFSET                      */
#include <asm/pgtable.h>            /* pte bit definitions              */

#if !defined (KERNEL_2_2)
#include <linux/spinlock.h>
#include <asm/semaphore.h>
#include <linux/highmem.h>
#else
#include <asm/spinlock.h>
#include <asm/semaphore.h>
#endif

#ifdef CONFIG_PROC_FS
#include <linux/proc_fs.h>
#endif

#ifdef CONFIG_DEVFS_FS
#include <linux/devfs_fs_kernel.h>
#endif

#ifdef CONFIG_KMOD
#include <linux/kmod.h>
#endif

#ifdef CONFIG_PM
#include <linux/pm.h>
#endif

#ifdef CONFIG_MTRR
#include <asm/mtrr.h>
#endif

#ifdef CONFIG_KDB
#include <linux/kdb.h>
#include <asm/kdb.h>
#endif

#if defined (CONFIG_AGP) || defined (CONFIG_AGP_MODULE)
#ifndef NOAGPGART
#  define AGPGART
#  include <linux/agp_backend.h>
#  include <linux/agpgart.h>
#endif
#endif

#if !defined (list_for_each)
#define list_for_each(pos, head) \
        for (pos = (head)->next; pos != (head); pos = (pos)->next)
#endif


/* various memory tracking/debugging techniques
 * disabled for retail builds, enabled for debug builds
 */

#ifdef DEBUG
#define NV_ENABLE_MEM_TRACKING 1
#endif

#if NV_ENABLE_MEM_TRACKING
#define NV_MEM_TRACKING_PAD_SIZE(size)   ((size) += sizeof(void *))
#define NV_MEM_TRACKING_HIDE_SIZE(ptr, size) \
    if ((ptr) && *(ptr)) { \
        *(int *) *(ptr) = (size); \
        (char *) *(ptr) += sizeof(void *); \
    }
#define NV_MEM_TRACKING_RETRIEVE_SIZE(ptr, size) \
    { \
        (char *) (ptr) -= sizeof(void *); \
        size = *(int *) (ptr); \
    }
#else
#define NV_MEM_TRACKING_PAD_SIZE(size)
#define NV_MEM_TRACKING_HIDE_SIZE(ptr, size)
#define NV_MEM_TRACKING_RETRIEVE_SIZE(ptr, size)  ((size) = 0)
#endif


/* poor man's memory allocation tracker.
 * main intention is just to see how much memory is being used to recognize
 * when memory usage gets out of control or if memory leaks are happening
 */

/* keep track of memory usage */
#if NV_ENABLE_MEM_TRACKING
extern int vm_usage;
extern int km_usage;
extern int fp_usage;

/* print out a running tally of memory allocation amounts, disabled by default */
// #define POOR_MANS_MEM_CHECK 1


/* slightly more advanced memory allocation tracker.
 * track who's allocating memory and print out a list of currently allocated
 * memory at key points in the driver
 */

#define MEMDBG_ALLOC(a,b) (a = kmalloc(b, GFP_ATOMIC))
#define MEMDBG_FREE(a)    (kfree(a))

#include "nv-memdbg.h"

#undef MEMDBG_ALLOC
#undef MEMDBG_FREE

extern struct mem_track_t *vm_list;
extern struct mem_track_t *km_list;
extern struct mem_track_t *fp_list;

/* print out list of memory allocations */
/* default to enabled for now */
#define LIST_MEM_CHECK 1

/* decide which memory types to apply mem trackers to */
#define VM_CHECKER 1
#define KM_CHECKER 1
#define FP_CHECKER 1

#endif  /* NV_ENABLE_MEM_TRACKING */

#define VM_PRINT(str, args...)
#define KM_PRINT(str, args...)
#define FP_PRINT(str, args...)

#define VM_ADD_MEM(a,b,c,d)
#define VM_FREE_MEM(a,b,c,d)
#define KM_ADD_MEM(a,b,c,d)
#define KM_FREE_MEM(a,b,c,d)
#define FP_ADD_MEM(a,b,c,d)
#define FP_FREE_MEM(a,b,c,d)

#define VM_ALLOC_RECORD(a,b,c)
#define VM_FREE_RECORD(a,b,c)
#define KM_ALLOC_RECORD(a,b,c)
#define KM_FREE_RECORD(a,b,c)


#if defined(VM_CHECKER)
#  if defined(POOR_MANS_MEM_CHECK)
#    undef  VM_PRINT
#    define VM_PRINT(str, args...)   printk(str, ##args)
#  endif
#  if defined(LIST_MEM_CHECK)
#    undef  VM_ADD_MEM
#    define VM_ADD_MEM(a,b,c,d)      nv_add_mem(&vm_list, a, b, c, d)
#    undef  VM_FREE_MEM
#    define VM_FREE_MEM(a,b,c,d)     nv_free_mem(&vm_list, a, b, c, d)
#  endif
#  undef VM_ALLOC_RECORD
#  define VM_ALLOC_RECORD(ptr, size, name) \
      if (ptr) { \
            vm_usage += size; \
            VM_PRINT("%s (0x%x: 0x%x): vm_usage is now 0x%x\n", \
                name, ptr, size, vm_usage); \
            VM_ADD_MEM(ptr, size, __FILE__, __LINE__); \
        }
#  undef VM_FREE_RECORD
#  define VM_FREE_RECORD(ptr, size, name) \
        if (ptr) { \
            vm_usage -= size; \
            VM_PRINT("%s (0x%x: 0x%x): vm_usage is now 0x%x\n", \
                name, ptr, size, vm_usage); \
            VM_FREE_MEM(ptr, size, __FILE__, __LINE__); \
        }
#endif

#if defined(KM_CHECKER)
#  if defined(POOR_MANS_MEM_CHECK)
#    undef  KM_PRINT
#    define KM_PRINT(str, args...)   printk(str, ##args)
#  endif
#  if defined(LIST_MEM_CHECK)
#    undef  KM_ADD_MEM
#    define KM_ADD_MEM(a,b,c,d)      nv_add_mem(&km_list, a, b, c, d)
#    undef  KM_FREE_MEM
#    define KM_FREE_MEM(a,b,c,d)     nv_free_mem(&km_list, a, b, c, d)
#  endif
#  undef KM_ALLOC_RECORD
#  define KM_ALLOC_RECORD(ptr, size, name) \
      if (ptr) { \
            km_usage += size; \
            KM_PRINT("%s (0x%x: 0x%x): km_usage is now 0x%x\n", \
                name, ptr, size, km_usage); \
            KM_ADD_MEM(ptr, size, __FILE__, __LINE__); \
        }
#  undef KM_FREE_RECORD
#  define KM_FREE_RECORD(ptr, size, name) \
        if (ptr) { \
            km_usage -= size; \
            KM_PRINT("%s (0x%x: 0x%x): km_usage is now 0x%x\n", \
                name, ptr, size, km_usage); \
            KM_FREE_MEM(ptr, size, __FILE__, __LINE__); \
        }
#endif

#if defined(FP_CHECKER)
#  if defined(POOR_MANS_MEM_CHECK)
#    undef  FP_PRINT
#    define FP_PRINT(str, args...)   printk(str, ##args)
#  endif
#if 0  // I had some problems tracking fp mem, so disable for now
#  if defined(LIST_MEM_CHECK)
#    undef  FP_ADD_MEM
#    define FP_ADD_MEM(a,b,c,d)      nv_add_mem(&fp_list, a, b, c, d)
#    undef  FP_FREE_MEM
#    define FP_FREE_MEM(a,b,c,d)     nv_free_mem(&fp_list, a, b, c, d)
#  endif
#endif
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 2, 18)
#define vmalloc_32 vmalloc
#endif

#define NV_VMALLOC(ptr, size) \
    { \
        (void *) (ptr) = vmalloc_32(size); \
        VM_ALLOC_RECORD(ptr, size, "vm_alloc"); \
    }

#define NV_VFREE(ptr, size) \
    { \
        VM_FREE_RECORD(ptr, size, "vm_alloc"); \
        vfree((void *) (ptr)); \
    }

#define NV_IOREMAP(ptr, physaddr, size) \
    { \
        (void *) (ptr) = ioremap(physaddr, size); \
        VM_ALLOC_RECORD(ptr, size, "vm_ioremap"); \
    }

#define NV_IOREMAP_NOCACHE(ptr, physaddr, size) \
    { \
        (void *) (ptr) = ioremap_nocache(physaddr, size); \
        VM_ALLOC_RECORD(ptr, size, "vm_ioremap_nocache"); \
    }

#define NV_IOUNMAP(ptr, size) \
    { \
        VM_FREE_RECORD(ptr, size, "vm_iounmap"); \
        iounmap(ptr); \
    }

/* only use this because GFP_KERNEL may sleep..
 * GFP_ATOMIC is ok, it won't sleep
 */
#define NV_KMALLOC(ptr, size) \
    { \
        (void *) (ptr) = kmalloc(size, GFP_KERNEL); \
        KM_ALLOC_RECORD(ptr, size, "km_alloc"); \
    }

#define NV_KMALLOC_ATOMIC(ptr, size) \
    { \
        (void *) (ptr) = kmalloc(size, GFP_ATOMIC); \
        KM_ALLOC_RECORD(ptr, size, "km_alloc_atomic"); \
    }  


#define NV_KFREE(ptr, size) \
    { \
        KM_FREE_RECORD(ptr, size, "km_free"); \
        kfree((void *) (ptr)); \
    }

#define NV_GET_FREE_PAGES(ptr, order) \
    { \
        (void *) (ptr) = __get_free_pages(NV_GFP_HW, order); \
    }
        
#define NV_FREE_PAGES(ptr, order) \
    { \
        free_pages(ptr, order); \
    }


#if !defined (KERNEL_2_2)
#  define LINUX_VMA_OFFS(vma)           (((vma)->vm_pgoff) << PAGE_SHIFT)
#  define GET_MODULE_SYMBOL(mod,sym)	(const void *) inter_module_get(sym)
#  define PUT_MODULE_SYMBOL(sym)        inter_module_put((char *) sym)
#  define GET_MAP_NR(phys_page)         virt_to_page(__va(phys_page))
#  define MEM_MAP_READ_COUNT(map_nr)    (atomic_read(&(map_nr)->count))
#  define MEM_MAP_INC_COUNT(map_nr)     (atomic_inc(&(map_nr)->count))
#  define MEM_MAP_DEC_COUNT(map_nr)     (atomic_dec(&(map_nr)->count))
#  define EVENT_QUEUE(ptr)               ((struct __wait_queue_head *)(ptr))
#  define VMA_PRIVATE(vma)              ((vma)->vm_private_data)
#  define INIT_EVENT_QUEUE(ptr) \
     if (ptr == NULL) { \
        NV_KMALLOC((ptr), sizeof(struct __wait_queue_head)); \
        if (ptr) { \
            memset((ptr), 0, sizeof(struct __wait_queue_head)); \
            init_waitqueue_head((struct __wait_queue_head *)(ptr)); \
        } \
     }
#  define FREE_EVENT_QUEUE(ptr)         NV_KFREE((ptr), sizeof(struct __wait_queue_head)); (ptr) = NULL;
#else
#  define in_irq()                      (local_irq_count[smp_processor_id()])
#  define LINUX_VMA_OFFS(vma)           ((vma)->vm_offset)
#  define GET_MODULE_SYMBOL(mod, sym)   (void*) get_module_symbol((mod), (sym))
#  define PUT_MODULE_SYMBOL(sym)
#  define GET_MAP_NR(phys_page)         MAP_NR(__va(phys_page))
#  define MEM_MAP_READ_COUNT(map_nr)    (atomic_read(&mem_map[map_nr].count))
#  define MEM_MAP_INC_COUNT(map_nr)     (atomic_inc(&mem_map[map_nr].count))
#  define MEM_MAP_DEC_COUNT(map_nr)     (atomic_dec(&mem_map[map_nr].count))
#  define EVENT_QUEUE(ptr)              ((struct wait_queue **) &(ptr))
#  define VMA_PRIVATE(vma)              ((void*)((vma)->vm_pte))
#  define INIT_EVENT_QUEUE(ptr)         if (ptr) { ((ptr) = NULL); }
#  define FREE_EVENT_QUEUE(ptr)
#endif

#if defined(REMAP_PAGE_RANGE_5)
#define NV_REMAP_PAGE_RANGE(a, b...)    remap_page_range(vma, a, ## b)
#elif defined(REMAP_PAGE_RANGE_4)
#define NV_REMAP_PAGE_RANGE(a, b...)    remap_page_range(a, ## b)
#else
#error "Couldn't determine number of arguments expected by remap_page_range!"
#endif

#if defined(pte_offset_atomic)
#define NV_PTE_OFFSET(addres, pg_mid_dir, pte) \
    { \
        pte_t *pte__ = pte_offset_atomic(pg_mid_dir, address); \
        pte = *pte__; \
        pte_kunmap(pte__); \
    }
#elif defined(pte_offset)
#define NV_PTE_OFFSET(addres, pg_mid_dir, pte) \
    pte = *pte_offset(pg_mid_dir, address)
#else
#define NV_PTE_OFFSET(addres, pg_mid_dir, pte) \
    { \
        pte_t *pte__ = pte_offset_map(pg_mid_dir, address); \
        pte = *pte__; \
        pte_unmap(pte__); \
    }
#endif

#define NV_PAGE_ALIGN(addr)             ( ((addr) + PAGE_SIZE - 1) / PAGE_SIZE)
#define NV_MASK_OFFSET(addr)            ( (addr) & (PAGE_SIZE - 1) )

#ifndef MAXMEM  /* temporary define for 2.2 kernels */
#define MAXMEM  (-PAGE_OFFSET - (64 * 1024 * 1024))
#endif

#ifndef NV01_ROOT
#define NV01_ROOT 0x00000000
#endif 

#if defined(NVCPU_IA64)
#define NV_GFP_HW (GFP_KERNEL | __GFP_DMA)
#define CACHE_FLUSH()
#else
#define NV_GFP_HW (GFP_KERNEL)
#define CACHE_FLUSH()  asm volatile("wbinvd":::"memory")
#endif

#if defined(NVCPU_X86) || defined(NVCPU_X86_64)
/* this isn't defined in some older 2.2 kernel header files */
#define NV_CPU_INTERRUPT_FLAGS_BIT (1<<9)
#elif defined(NVCPU_IA64)
/* For whatever reason this is not defined an any header file I could
 * find.  From Intel IA64 Architecture Software Developers Manual Volume 2: 
 * IA64 System Architecture page 3-7 we have:
 */
#define NV_CPU_INTERRUPT_FLAGS_BIT (1<<14)
#else
#error define NV_CPU_INTERRUPT_FLAGS_BIT
#endif

/*
 * AMD Athlon processors expose a subtle bug in the Linux
 * kernel, that may lead to AGP memory corruption. Recent
 * kernel versions had a workaround for this problem, but
 * 2.4.20 is the first kernel to address it properly. The
 * page_attr API provides the means to solve the problem. 
 */
#if defined(NVCPU_X86) && (LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,20))
static inline void NV_SET_AMD_PAGE_ATTRIB(void **page_list, int count)
    {
        int i;
        for (i = 0; i < count; i++)
        {
            struct page *page = virt_to_page(__va(page_list[i]));
            change_page_attr(page, 1, PAGE_KERNEL_NOCACHE);
        }
    }
static inline void NV_CLEAR_AMD_PAGE_ATTRIB(void **page_list, int count)
    {
        int i;
        for (i = 0; i < count; i++)
        {
            struct page *page = virt_to_page(__va(page_list[i]));
            change_page_attr(page, 1, PAGE_KERNEL);
        }
    }
#else
#define NV_SET_AMD_PAGE_ATTRIB(page_list, count)
#define NV_CLEAR_AMD_PAGE_ATTRIB(page_list, count)
#endif

static inline int NV_IRQL_IS_RAISED()
    {
        unsigned long int eflags;
        __save_flags(eflags);
        return !(eflags & NV_CPU_INTERRUPT_FLAGS_BIT);
    }
 
static inline int calc_order(size)
    {
        int order = 0;
        while ( ((1 << order) * PAGE_SIZE) < (size))
        {
            order++;
        }
        return order;
    }

#if !defined (pgprot_noncached)
static inline pgprot_t pgprot_noncached(pgprot_t old_prot)
    {
        pgprot_t new_prot = old_prot;
        if (boot_cpu_data.x86 > 3)
            new_prot = __pgprot(pgprot_val(old_prot) | _PAGE_PCD);
        return new_prot;
    }
#endif

/*
 * An allocated bit of memory using NV_MEMORY_ALLOCATION_OFFSET
 *   looks like this in the driver
 */

typedef struct nv_alloc_s {
    struct nv_alloc_s *next;    
    struct vm_area_struct *vma;
    unsigned int   usage_count;
    unsigned int   flags;
    unsigned int   agp_offset;
    unsigned int   num_pages;
    void         **page_table;          /* list of physical pages allocated */
    void          *key_mapping;         /* mapping used as a key for finding this nv_alloc_t */
                                        /*   may be the same as page_table                   */
    unsigned int   class;
    void          *priv_data;
} nv_alloc_t;

#define NV_ALLOC_TYPE_PCI      (1<<0)
#define NV_ALLOC_TYPE_AGP      (1<<1)
#define NV_ALLOC_TYPE_CONTIG   (1<<2)
#define NV_ALLOC_TYPE_KERNEL   (1<<3)
#define NV_ALLOC_TYPE_VMALLOC  (1<<4)


/* linux-specific version of old nv_state_t */
/* this is a general os-specific state structure. the first element *must* be
   the general state structure, for the generic unix-based code */
typedef struct {
    nv_state_t nv_state;

    nv_alloc_t *alloc_queue;

    // bottom half interrupt handler info; per device
    /* keep track of any pending bottom-halves */
    struct tq_struct *bh;
    atomic_t bh_count;

    /* get a timer callback every second */
    struct timer_list rc_timer;

    /* per-device locking mechanism for access to core rm */
    spinlock_t rm_lock;
    int rm_lock_cpu;
    int rm_lock_count;

    /* lock for linux-specific data, not used by core rm */
    /* XXX eventually change to sleeping locks? */
    spinlock_t ldata_lock;

    /* lock for linux-specific alloc queue */
    /* XXX eventually change to sleeping locks? */
    spinlock_t at_lock;
} nv_linux_state_t;


/*
 * file-private data
 * hide a pointer to our data structures in a file-private ptr
 * there are times we need to grab this data back from the file
 * data structure..
 */

#define NV_EVENT_FIFO_SIZE 6

typedef struct
{
    void *nvptr;
    U032 num_events;
    U032 put, get;
    spinlock_t fp_lock;
    void *wqueue;               // wait_queue for polling
    nv_event_t *event_fifo;     // fifo for storing events
} nv_file_private_t;

#define FILE_PRIVATE(filep)     ((filep)->private_data)

#define NV_GET_NVFP(filep)      ((nv_file_private_t *) FILE_PRIVATE(filep))

/* for the card devices */
#define NVL_FROM_FILEP(filep) \
    ((nv_linux_state_t*) (NV_GET_NVFP(filep))->nvptr)

#define NV_GET_NVL_FROM_NV_STATE(nv) \
    ((nv_linux_state_t *) nv->os_state)

#define NV_STATE_PTR(nvl)   (&((nvl)->nv_state))

#define NV_ATOMIC_INC(data)  atomic_inc(&(data))
#define NV_ATOMIC_DEC(data)  atomic_dec(&(data))
#define NV_ATOMIC_READ(data) atomic_read(&(data))

/*
 * Basic support for kgdb assertions.
 */
#if defined(CONFIG_X86_REMOTE_DEBUG)
#include <linux/gdb.h>

#define NV_ASSERT(message, condition) KGDB_ASSERT(message, condition)
#else
#define NV_ASSERT(message, condition)
#endif


#endif  /* _NV_LINUX_H_ */
